"""Includes ObjectClfSettings class"""
from typing import Tuple

from organon.ml.common.enums.classification_type import ClassificationType
from organon.ml.common.enums.color_type import ColorType
from organon.ml.object_classification.domain.objects.fine_tuning_settings import FineTuningSettings
from organon.ml.object_classification.domain.objects.pretrained_model_settings import PretrainedModelSettings
from organon.ml.object_classification.domain.objects.transfer_learning_settings import TransferLearningSettings
from organon.ml.object_classification.domain.objects.data_augmentation_settings import DataAugmentationSettings


class ObjectClfSettings:
    """Settings for object classification"""

    def __init__(self, train_data_dir: str,
                 pretrained_model_settings: PretrainedModelSettings = None,
                 transfer_learning_settings: TransferLearningSettings = None,
                 fine_tuning_settings: FineTuningSettings = None,
                 data_augmentation_settings: DataAugmentationSettings = None,
                 validation_data_dir: str = None,
                 validation_data_ratio: float = 0.2, image_size: Tuple[int, int] = (150, 150),
                 batch_size: int = 50, prediction_threshold: float = 0, color_mode: ColorType = ColorType.RGB,
                 random_seed: int = 42):
        # pylint: disable=too-many-arguments
        self.train_data_dir: str = train_data_dir
        self.validation_data_dir: str = validation_data_dir
        self.clf_mode: ClassificationType = None
        self.color_mode: ColorType = color_mode
        self.validation_data_ratio: float = validation_data_ratio
        self.image_size: Tuple[int, int] = image_size
        self.batch_size: int = batch_size
        self.pretrained_model_settings = pretrained_model_settings if pretrained_model_settings is not None \
            else PretrainedModelSettings()
        self.transfer_learning_settings = transfer_learning_settings if transfer_learning_settings is not None \
            else TransferLearningSettings()
        self.data_augmentation_settings = data_augmentation_settings
        self.fine_tuning_settings = fine_tuning_settings
        self.prediction_threshold = prediction_threshold
        self.random_seed: int = random_seed



"""
This module includes ExecutionPlanService class
"""
from typing import Dict

from organon.afe.domain.common.runtime_stats_service import RuntimeStatsService
from organon.afe.domain.modelling.businessobjects.target_file_record_collection import TargetFileRecordCollection
from organon.afe.domain.modelling.businessobjects.transaction_file_record_collection import \
    TransactionFileRecordCollection
from organon.afe.domain.settings.afe_algorithm_settings import AfeAlgorithmSettings
from organon.afe.domain.settings.feature_generation_settings import FeatureGenerationSettings
from organon.afe.domain.settings.record_source import RecordSource


# pylint: disable=no-self-use
# pylint: disable=unused-argument

class BaseExecutionPlanService:
    """
    Class for estimating remaining time and plan of the service
    """
    GETTING_TARGETS_KEY = 'Getting Targets'
    DECIDE_DIM_QTY_COL_KEY = 'Deciding dim and qty columns'
    CREATE_DISTINCT_ENTITIES_KEY = 'Creating dist. entities table'
    READ_TRX_FILE_KEY = 'Reading trx file'
    REDUCE_COLUMNS_KEY = 'Reducing columns'
    EXTRACT_ALL_VARS_KEY = 'Extracting all variables'
    OUTPUT_KEY = 'Output'

    def __init__(self):
        """
        Initializes object with settings
        """
        self.iteration_count = None
        self.total_reduction_loop = None
        self._runtime_stats_service = RuntimeStatsService()

    def get_runtime_statistics(self):
        """returns recorded runtime statistics"""
        return self._runtime_stats_service.runtime_statistics

    def start_recording_runtime_stats(self, interval_in_seconds: float = 1):
        """Starts recording runtime statistics"""
        self._runtime_stats_service.start_recording_runtime_stats(interval_in_seconds)

    def stop_recording_runtime_stats(self):
        """Stops recording runtime statistics"""
        self._runtime_stats_service.stop_recording_runtime_stats()

    def add_df_builder_statistic(self, target_count, trx_count, d_count, date_col_name, generated_feature, exec_time):
        """
        Adds df_builder statistics
        :param target_count: target count
        :param trx_count: trx count
        :param d_count: dimension count
        :param date_col_name: name of date column corresponding to current statistic
        :param generated_feature: generated features
        :param exec_time: time spent
        """

    def add_unsupervised_statistic(self, column_count, bin_count, record_count, r_factor, selected_count, exec_time):
        """
        Adds unsupervised statistics
        :param column_count: column count
        :param bin_count: bin count
        :param record_count: record count
        :param r_factor: r factor
        :param selected_count: selected count
        :param exec_time: time spent
        """

    def add_supervised_statistic(self, record_count, column_count, exec_time):
        """
        Adds Supervised model statistics
        :param record_count: record count
        :param column_count: column count
        :param exec_time: time past
        """

    def save_statistics(self):
        """
        Saves statistics according to type
        """

    def set_collections(self,
                        trx_collection: TransactionFileRecordCollection,
                        records_per_file: Dict[RecordSource, TargetFileRecordCollection]):
        """
        Sets collections and initializes estimations
        :param trx_collection: trx collections
        :param records_per_file: targets
        """

    def start_get_targets(self):
        """Add start time for reading targets to runtime statistics as event"""
        self._add_event_to_runtime_statistics(self.GETTING_TARGETS_KEY)

    def complete_get_targets(self, exec_time):
        """
        Moves get_targets from to_do_map to completed map with time
        :param exec_time: time past
        """

    def start_create_distinct_entities_table(self):
        """Add start time for creating distinct entities table to runtime statistics as event"""

    def complete_create_distinct_entities_table(self, exec_time):
        """
        Moves create_distinct_entities_table from to_do_map to completed map with time
        :param exec_time: time past
        """

    def start_decide_columns(self):
        """Add start time for deciding dim-qty columns to runtime statistics as event"""
        self._add_event_to_runtime_statistics(self.DECIDE_DIM_QTY_COL_KEY)

    def complete_decide_columns(self, exec_time):
        """
        Moves decide_columns from to_do_map to completed map with time
        :param exec_time: time past
        """

    def start_trx_file(self):
        """Add start time for reading transaction file to runtime statistics as event"""
        self._add_event_to_runtime_statistics(self.READ_TRX_FILE_KEY)

    def complete_trx_file(self, exec_time):
        """
        Moves trx_file from to_do_map to completed map with time
        :param exec_time: time past
        """

    def start_reduce(self):
        """Add start time for reducing columns to runtime statistics as event"""
        self._add_event_to_runtime_statistics(self.REDUCE_COLUMNS_KEY)

    def complete_reduce(self, exec_time):
        """
        Moves reduce from to_do_map to completed map with time
        :param exec_time: time past
        """

    def start_extracting(self):
        """Add start time for extracting all variables to runtime statistics as event"""

    def complete_extracting(self, exec_time):
        """
        Moves extracting from to_do_map to completed map with time
        :param exec_time: time past
        """

    def start_output(self):
        """Add start time for generating/writing outputs to runtime statistics as event"""
        self._add_event_to_runtime_statistics(self.OUTPUT_KEY)

    def complete_output(self, exec_time):
        """
        Moves output from to_do_map to completed map with time
        :param exec_time: time past
        """

    def calculate_reduce_time(self):
        """
        Calculates the total time for reduce
        """

    def calculate_calibrated_reduce_time(self, record_source, order: int,
                                         dimension_column, df_time, model_time):
        """
        Returns calibrated estimation for reduce
        :param record_source: current target source
        :param dimension_column: current dimension
        :param df_time: time past for current iteration
        :param model_time: time past for current iteration
        :return: calibrated time for rest of the iterations
        """

    def get_plan(self):
        """
        Returns the plan

        :return: returns the plan
        """
        return None

    @staticmethod
    def log_trx_remaining(time_past: float, record_fetched: int, fetch_failure: int, record_count: int):
        """
        Logs the remaining info of fetching trx
        :param time_past: time past until log
        :param record_fetched: record fetched successfully
        :param fetch_failure: record failed
        :param record_count: total records
        """

    @staticmethod
    def get_column_count_per_dimension(algorithm_settings: AfeAlgorithmSettings,
                                       feature_generation_setting: FeatureGenerationSettings):
        """
        Returns estimated columns to be generated per dimension
        :param feature_generation_setting:
        :param algorithm_settings: settings
        :return: estimation per dimension
        """

    def _add_event_to_runtime_statistics(self, key: str):
        self._runtime_stats_service.add_event(key)



"""Includes BaseModelGenerator class"""
from typing import Tuple, TYPE_CHECKING
from organon.ml.common.enums.pretrained_model_type import PretrainedModelType
from organon.ml.common.enums.color_type import ColorType
from organon.ml.common.enums.random_flip_type import RandomFlipType
from organon.ml.object_classification.common.helpers import throw_val_ex_with_log
from organon.ml.object_classification.domain.objects.data_augmentation_settings import DataAugmentationSettings
from organon.ml.object_classification.domain.objects.pretrained_model_settings import PretrainedModelSettings

if TYPE_CHECKING:
    import tensorflow as tf  # noqa


class BaseModelGenerator:
    """BaseModelGenarator class"""

    @staticmethod
    def get_base_model(image_size: Tuple[int, int], color_mode: ColorType,
                       pretrained_model_settings: PretrainedModelSettings,
                       data_augmentation_settings: DataAugmentationSettings = None, random_seed: int = 42) -> \
            Tuple["tf.Tensor", "tf.keras.Input"]:
        """Returns base_model with given settings"""
        import tensorflow as tf  # pylint: disable=import-outside-toplevel

        input_shape = image_size + BaseModelGenerator._get_input_dim(color_mode)

        inputs = tf.keras.Input(shape=input_shape)

        if data_augmentation_settings is not None:
            data_augmentation = BaseModelGenerator._get_augmentation_layers(data_augmentation_settings,
                                                                            random_seed=random_seed)
            inputs = data_augmentation(inputs)

        base_model_args = dict(weights=pretrained_model_settings.weights,
                               input_shape=input_shape, include_top=False)
        base_model_type = pretrained_model_settings.model

        if base_model_type == PretrainedModelType.XCEPTION:
            base_model = tf.keras.applications.Xception(**base_model_args)
            process_input = tf.keras.applications.xception.preprocess_input(inputs)
        elif base_model_type == PretrainedModelType.INCEPTION_RES_NET_V2:
            base_model = tf.keras.applications.InceptionResNetV2(**base_model_args)
            process_input = tf.keras.applications.inception_resnet_v2.preprocess_input(inputs)
        elif base_model_type == PretrainedModelType.RES_NET_50_V2:
            base_model = tf.keras.applications.resnet_v2.ResNet50V2(**base_model_args)
            process_input = tf.keras.applications.resnet_v2.preprocess_input(inputs)
        elif base_model_type == PretrainedModelType.EFFICIENT_NET_B4:
            base_model = tf.keras.applications.efficientnet.EfficientNetB4(**base_model_args)
            process_input = tf.keras.applications.efficientnet.preprocess_input(inputs)
        elif base_model_type == PretrainedModelType.EFFICIENT_NET_B7:
            base_model = tf.keras.applications.efficientnet.EfficientNetB7(**base_model_args)
            process_input = tf.keras.applications.efficientnet.preprocess_input(inputs)
        else:
            return throw_val_ex_with_log("Model type is not defined!")

        base_model.trainable = False

        base_model = base_model(process_input, training=False)

        return base_model, inputs

    @staticmethod
    def _get_augmentation_layers(data_augmentation_settings: DataAugmentationSettings,
                                 random_seed: int) -> "tf.keras.Sequential":
        import tensorflow as tf  # pylint: disable=import-outside-toplevel
        if data_augmentation_settings.random_flip == RandomFlipType.HORIZONTAL_AND_VERTICAL:
            random_flip_mode = "horizontal_and_vertical"
        elif data_augmentation_settings.random_flip == RandomFlipType.HORIZONTAL:
            random_flip_mode = "horizontal"
        elif data_augmentation_settings.random_flip == RandomFlipType.VERTICAL:
            random_flip_mode = "vertical"
        else:
            return throw_val_ex_with_log("Random flip mode is ambiguous")

        seed = random_seed
        return tf.keras.Sequential([
            tf.keras.layers.RandomFlip(random_flip_mode, seed=seed),
            tf.keras.layers.RandomRotation(data_augmentation_settings.random_rotation, seed=seed),
            tf.keras.layers.RandomZoom(data_augmentation_settings.random_zoom, seed=seed)])

    @staticmethod
    def _get_input_dim(color_mode: ColorType) -> tuple:
        if color_mode == ColorType.RGB:
            input_dim = (3,)
        elif color_mode == ColorType.RGBA:
            input_dim = (4,)
        elif color_mode == ColorType.GRAY:
            input_dim = (1,)
        else:
            return throw_val_ex_with_log("Color type is ambiguous!")
        return input_dim



"""Includes helper functions to sample, manipulate dataframes"""
from typing import List

import numpy as np
import pandas as pd

from organon.fl.core.helpers import list_helper


def get_train_test_split(data: pd.DataFrame, test_ratio: float, strata_columns: List[str] = None, random_state=None):
    """Divides given data to two splits(train and test) using given ratio"""
    if test_ratio > 1.0:
        raise ValueError("test_ratio cannot be higher than 1.0")
    test_indices = get_sample_indices(data, test_ratio, strata_columns=strata_columns, random_state=random_state)
    train_data = data.loc[data.index.difference(test_indices)]
    return train_data, data.loc[test_indices]


def get_train_test_split_from_strata(data: pd.DataFrame, strata_data: pd.Series, test_ratio: float, random_state=None):
    """Divides given data to two splits(train and test) using given ratio"""
    if test_ratio > 1.0:
        raise ValueError("test_ratio cannot be higher than 1.0")
    test_indices = get_sample_indices_from_series(strata_data, test_ratio, random_state=random_state)
    return data.loc[data.index.difference(test_indices)], strata_data.loc[strata_data.index.difference(test_indices)],\
           data.loc[test_indices], strata_data.loc[test_indices]


def get_sample_indices_from_series(strata_data: pd.Series, frac: float, replace=False, random_state=None):
    """
    Samples data and returns indices of selected rows.
    """
    return strata_data.groupby(strata_data, group_keys=False).sample(frac=frac, replace=replace,
                                                                    random_state=random_state).index


def get_sample_indices(data: pd.DataFrame, frac: float, strata_columns: List[str] = None, replace=False,
                       random_state=None):
    """
    Samples data and returns indices of selected rows.
    """
    if not list_helper.is_null_or_empty(strata_columns):
        if len([col for col in data.columns if col not in strata_columns]) > 0:
            data = data[strata_columns]
        indices = data.groupby(strata_columns, group_keys=False).sample(frac=frac, replace=replace,
                                                                        random_state=random_state).index
    else:
        indices = data.sample(frac=frac, replace=replace, random_state=random_state).index
    return indices


def get_sample_data(data: pd.DataFrame, frac: float, strata_columns: List[str] = None, replace=False,
                    random_state=None):
    """Samples data and returns new dataframe"""
    return data.loc[
        get_sample_indices(data, frac, strata_columns=strata_columns, replace=replace, random_state=random_state)]


def get_bins(data: pd.Series, num_bins: int) -> pd.Series:
    """Splits given data into given number of bins and returns bin numbers for every row"""
    sorted_indices = data.sort_values().index
    len_index = len(sorted_indices)
    num_rows_per_bin = int(len_index / num_bins)
    remaining = len_index % num_bins
    arr = np.empty(len_index)
    curr_index = 0
    for i in range(num_bins):
        step = num_rows_per_bin + 1 if i < remaining else num_rows_per_bin
        arr[curr_index:curr_index + step] = i
        curr_index = curr_index + step

    return pd.Series(arr, index=sorted_indices, dtype="category").astype(np.int8).sort_index()



"""Includes Sampler class."""
from typing import List

import pandas as pd

from organon.ml.sampling.domain.objects.sampling_output import SamplingOutput
from organon.ml.sampling.domain.services.sampling_service import SamplingService
from organon.ml.sampling.services.user_settings.user_sampling_settings import UserSamplingSettings
from organon.ml.sampling.settings.sampling_user_input_service import SamplingUserInputService


class Sampler:
    """User Interface class for sampling"""

    def __init__(self):
        self._output: SamplingOutput = None

    def execute(self, data: pd.DataFrame, sampling_strategy: str = None, target_column_name: str = None,
                target_type: str = None, strata_columns: List[str] = None, split_test_data: bool = None,
                test_split_ratio: float = None, sampling_ratio: float = None, data_sample_ratio: float = None):
        """todo"""
        # pylint: disable=too-many-arguments
        user_settings = UserSamplingSettings(data, target_column_name, target_type, strata_columns, split_test_data,
                                             test_split_ratio, sampling_strategy, sampling_ratio, data_sample_ratio)
        settings = SamplingUserInputService.get_sampling_settings(user_settings)
        output = SamplingService.sample(settings)
        self._output = output
        return output



"""Includes MultiClassReporterSettings class."""
from dataclasses import dataclass
from typing import List

import numpy as np

from organon.ml.reporting.settings.objects.reporter_settings import ReporterSettings


@dataclass
class MultiClassReporterSettings(ReporterSettings):
    """Settings for MultiClass Reporter service"""
    probability_values: np.array = None
    ordered_class_names: List = None



"""Includes DqStatisticalFunctions class."""
from typing import Dict, Tuple, TypeVar

import numpy as np

from organon.fl.core.collections.sorted_dict import SortedDict
from organon.fl.mathematics.businessobjects.sets.interval import Interval
from organon.fl.mathematics.businessobjects.sets.sum_triple import SumTriple
from organon.fl.mathematics.constants import DOUBLE_MIN, NEGATIVE_INFINITY, POSITIVE_INFINITY
from organon.fl.mathematics.enums.sets.interval_type import IntervalType

T = TypeVar("T")


class DqStatisticalFunctions:
    """Includes helper functions for dq statistics."""

    @staticmethod
    def get_percentages_from_frequencies_for_sorted(frequencies: SortedDict[object, float]) -> \
            SortedDict[object, float]:
        """todo"""
        total = sum(frequencies.values())
        return SortedDict({key: value / total for key, value in frequencies.items()})

    @staticmethod
    def get_percentages_from_frequencies(frequencies: Dict[object, float]) -> Dict[object, float]:
        """todo"""
        total = sum(frequencies.values())
        return {key: value / total for key, value in frequencies.items()}

    @staticmethod
    def get_balanced_intervals_and_statistics(frequencies: SortedDict[float, int],
                                              max_num_of_intervals: float,
                                              min_bucket_size: int) -> SortedDict[Interval, SumTriple]:
        """Get balanced intervals and statistics for interval_statistics of stats"""
        if frequencies is None:
            raise Exception("Frequencies is null as input to GetBalancedIntervals method")
        values_length = sum(frequencies.values())
        if values_length == 0:
            raise Exception("Frequencies is empty as input to GetBalancedIntervals method")
        if max_num_of_intervals <= 1:
            raise Exception(
                "Maximum-number-of-intervals argument must be greater than 1 as input to GetBalancedIntervals method")
        if values_length < max_num_of_intervals:
            raise Exception(
                "Maximum-number-of-intervals argument must be less than array length as"
                " input to GetBalancedIntervals method")
        window = int(float(values_length) / max_num_of_intervals)
        lower_bound = NEGATIVE_INFINITY
        cum = 0
        triple = SumTriple()
        interval_dict: SortedDict[Interval, SumTriple] = SortedDict()
        updated_bucket_size = int(max(window * 0.2, min_bucket_size))
        for order, (frequency_key, frequency_value) in enumerate(frequencies.items()):
            cum = cum + frequency_value
            triple.update_with_data(frequency_value, frequency_key)
            if order == (len(frequencies) - 1):
                if (len(interval_dict) > 0) & (cum < updated_bucket_size):
                    last_interval = interval_dict.keys()[len(interval_dict) - 1].deep_copy()
                    last_triple = interval_dict.values()[len(interval_dict) - 1]
                    modified_triple = SumTriple(last_triple.count, last_triple.sum_, last_triple.sum_of_squares)
                    modified_triple.update_with_sum_triple(triple)
                    interval_dict.popitem()
                    modified_interval = Interval("", last_interval.lower_bound, POSITIVE_INFINITY,
                                                 IntervalType.OPEN_CLOSED)
                    interval_dict[modified_interval] = modified_triple
                else:
                    interval = Interval("", lower_bound, POSITIVE_INFINITY, IntervalType.OPEN_CLOSED)
                    interval_dict[interval] = triple
            elif cum >= window:
                upper_bound = frequency_key
                interval = Interval("", lower_bound, upper_bound, IntervalType.OPEN_CLOSED)
                interval_dict[interval] = triple
                lower_bound = upper_bound
                cum = 0
                triple = SumTriple()
        return interval_dict

    @staticmethod
    def population_stability_index(left_frequency: Dict[T, float] or SortedDict[str, float],
                                   right_frequency: Dict[T, float] or SortedDict[str, float],
                                   minimum_injection: int) -> Tuple[T, float]:
        """Return population stability index value"""

        union_key = list(left_frequency.keys() | right_frequency.keys())
        updated_left_frequency = {}
        updated_right_frequency = {}
        for item in union_key:
            updated_left_frequency[item] = left_frequency[item] if (
                    item in left_frequency and left_frequency[item] > 0) else minimum_injection

            updated_right_frequency[item] = right_frequency[item] if (
                    item in right_frequency and right_frequency[item] > 0) else minimum_injection

        left_percentages = DqStatisticalFunctions.get_percentages_from_frequencies(updated_left_frequency)
        right_percentages = DqStatisticalFunctions.get_percentages_from_frequencies(updated_right_frequency)
        return DqStatisticalFunctions._get_population_stability_index(left_percentages, right_percentages)

    @staticmethod
    def _get_population_stability_index(left_percentages: Dict[T, float],
                                        right_percentages: Dict[T, float]) -> Tuple[T, float]:

        max_ = DOUBLE_MIN
        max_category = None
        sum_ = 0.0
        for item_key, item_value in left_percentages.items():
            left_percentage = item_value
            right_percentage = right_percentages[item_key]
            current = (left_percentage - right_percentage) * np.log(left_percentage / right_percentage)
            sum_ = sum_ + current
            if current >= max_:
                max_ = current
                max_category = item_key
        return max_category, sum_



"""This module includes IDataFrameOperations abstract class."""
import abc
from typing import Dict, Any, List

from organon.fl.core.businessobjects.idataframe import IDataFrame


class IDataFrameOperations(metaclass=abc.ABCMeta):
    """Abstract class for Dataframe Operations."""

    @staticmethod
    @abc.abstractmethod
    def join(df1: IDataFrame, df2: IDataFrame):
        """:returns pandas join function equivalent"""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_column_stability(df_obj: IDataFrame):
        """:returns boolean df depends on whether a column is stable or not"""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_column_distinct_counts(df_obj: IDataFrame, col_names: List[str] = None) -> Dict[str, int]:
        """:returns number of unique values per column"""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_min_value(df_obj: IDataFrame, col_name: str, skipna=True) -> float:
        """Returns min value in given column"""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_max_value(df_obj: IDataFrame, col_name: str, skipna=True) -> float:
        """Returns max value in given column"""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_mean_value(df_obj: IDataFrame, col_name: str, skipna=True) -> float:
        """Returns mean value in given column"""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_trimmed_mean_value(df_obj: IDataFrame, col_name: str, lower_bound: float, upper_bound: float) -> float:
        """Returns trimmed mean value in given column"""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_variance_value(df_obj: IDataFrame, col_name: str, skipna=True) -> float:
        """Returns variance in given column"""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_sum_value(df_obj: IDataFrame, col_name: str, skipna=True) -> float:
        """Returns sum of values in given column"""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_percentile_values(df_obj: IDataFrame, col_name: str, percentiles: List[float]) -> Dict[float, float]:
        """Returns percentile values for given percentiles"""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def get_frequencies(df_obj: IDataFrame, col_name: str, values: list = None) -> Dict[Any, int]:
        """Returns number of occurrences of given values in the given column"""

    @staticmethod
    @abc.abstractmethod
    def get_col_df_without_values(df_obj: IDataFrame, col_name: str, values: list) -> IDataFrame:
        """Returns the given column in a new DataFrame by eliminating given values"""

    @staticmethod
    @abc.abstractmethod
    def get_size(df_obj: IDataFrame):
        """Returns number of rows in dataframe"""

    @staticmethod
    @abc.abstractmethod
    def get_col_values(df_obj: IDataFrame, col_name: str) -> list:
        """Returns column values as list"""



"""
This module keeps the helper functions for guid type.
"""
import uuid


def new_guid(max_length: int = None) -> str:
    """
    Returns a uuid value as a string.
    Max_length parameter limits the maximum length of the returned string.
    :param max_length: maximum length of the uuid value
    :return: uuid value as a string
    """
    if max_length is None:
        return str(uuid.uuid4())
    if 0 < max_length < 36:
        return str(uuid.uuid4()).replace("-", "")[:max_length]
    return ""



"""Module for LinearRegressionParameter"""
from dataclasses import dataclass


@dataclass
class LinearRegressionParameter:
    """Statistical info for swept parameters"""
    estimate: float
    standard_error: float
    test_statistic: float
    p_value: float
    vif: float
    standardized_coefficient: float
    lower_confidence_limit: float
    upper_confidence_limit: float



"""Includes DqDfCalculationParameters class."""

from organon.idq.domain.settings.abstractions.dq_base_calculation_parameters import DqBaseCalculationParameters
from organon.idq.domain.settings.input_source.dq_df_input_source_settings import DqDfInputSourceSettings


class DqDfCalculationParameters(DqBaseCalculationParameters[DqDfInputSourceSettings]):
    """DQ calculation parameters for dataframe source type"""



"""Includes BaseUnsupervisedFeatureSelectionSettings class"""
from organon.ml.feature_selection.domain.objects.settings.base_feature_selection_settings import \
    BaseFeatureSelectionSettings


class BaseUnsupervisedFeatureSelectionSettings(BaseFeatureSelectionSettings):
    """Base settings for unsupervised feature selection services"""



"""Includes ModelType enum class"""
from enum import Enum, auto


class ModelTypes(Enum):
    """Ultralytics model types"""

    NANO = auto()
    SMALL = auto()
    MEDIUM = auto()
    LARGE = auto()
    XLARGE = auto()



"""
This module ColumnNativeType enum class.
"""
from enum import Enum


# pylint: disable=invalid-name
class ColumnNativeType(Enum):
    """
    Enum class for defining column native type values.
    """
    String = 1
    Numeric = 2
    Other = 3
    Date = 4



"""This module includes DataFrameBuilder class."""
import copy
import math
from typing import Dict, List

import numpy as np
import pandas as pd

from organon.afe.core.businessobjects.afe_static_objects import AfeStaticObjects
from organon.afe.dataaccess.services.transaction_file_record_reader import TransactionFileRecordReader
from organon.afe.domain.enums.afe_operator import AfeOperator, OPERATOR_SET_ONE, OPERATOR_SET_TWO
from organon.afe.domain.enums.afe_target_column_type import AfeTargetColumnType
from organon.afe.domain.enums.binary_target_class import BinaryTargetClass
from organon.afe.domain.enums.date_resolution import DateResolution
from organon.afe.domain.modelling.businessobjects.afe_column import AfeColumn
from organon.afe.domain.modelling.businessobjects.base_afe_feature import BaseAfeFeature
from organon.afe.domain.modelling.businessobjects.target_file_record_collection import TargetFileRecordCollection
from organon.afe.domain.modelling.businessobjects.transaction_file_record_collection import \
    TransactionFileRecordCollection
from organon.afe.domain.modelling.businessobjects.transaction_file_stats import TransactionFileStats
from organon.afe.domain.modelling.helper import data_frame_builder_helper  # pylint: disable=no-name-in-module
from organon.afe.domain.settings.afe_date_column import AfeDateColumn
from organon.afe.domain.settings.feature_generation_settings import FeatureGenerationSettings
from organon.afe.domain.settings.target_descriptor import TargetDescriptor
from organon.fl.core.businessobjects.histogram_16 import Histogram16
from organon.fl.core.businessobjects.np2d_dataframe import Np2dDataFrame
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.helpers import date_helper
from organon.fl.logging.helpers.log_helper import LogHelper


# pylint: disable=too-many-instance-attributes
class DataFrameBuilder:
    """Class for generating AFE features and building a dataframe which stores feature values for every entity."""

    operators = [AfeOperator.Sum, AfeOperator.Max, AfeOperator.Min, AfeOperator.Frequency,
                 AfeOperator.Density, AfeOperator.CountDistinct, AfeOperator.Mode,
                 AfeOperator.TimeSinceLast, AfeOperator.TimeSinceFirst, AfeOperator.Ratio]

    # pylint: disable=too-many-arguments
    def __init__(self,
                 target_descriptor: TargetDescriptor,
                 target_file_record_collection: TargetFileRecordCollection,
                 trx_file_record_collection: TransactionFileRecordCollection,
                 transaction_file_stats: TransactionFileStats = None,
                 dimensions_per_date_col: Dict[int, set] = None,
                 horizons_per_date_col: Dict[int, set] = None,
                 max_num_of_columns: int = None
                 ):

        self._target_descriptor = target_descriptor
        self._target_file_record_collection = target_file_record_collection
        self._trx_file_record_collection = trx_file_record_collection
        self._dimensions_per_date_col = dimensions_per_date_col
        self._horizons_per_date_col = horizons_per_date_col
        if transaction_file_stats is None:
            self._trx_stats = trx_file_record_collection.transaction_file_stats
        else:
            self._trx_stats = transaction_file_stats

        if max_num_of_columns is None:
            if dimensions_per_date_col is None or horizons_per_date_col is None is None:
                raise KnownException("Please supply dimensions_per_date_col, horizons_per_date_col"
                                     " to calculate max number of columns")
            max_num_of_columns = self._get_maximum_number_of_frame_columns()

        if target_file_record_collection is not None:
            self._frame = Np2dDataFrame(target_file_record_collection.sampled_count,
                                        max_num_of_columns, np.float32)

        self.target_array: np.array = None
        self._record_index_target_index_map: Dict[int, str] = {}
        self._name_to_column: Dict[str, AfeColumn] = {}
        self._group_to_name: Dict[str, List[str]] = {}
        self._index_to_id: Dict[int, str] = None
        self._index_to_date: Dict[int, int] = None
        self._included_operators = None
        self._included_operator_set_one = None
        self._included_operator_set_two = None
        self._new_col_names_map: Dict[str, str] = {}

    @property
    def frame(self):
        """Returns dataframe."""
        return self._frame

    @property
    def name_to_column(self):
        """Returns 'column name - AfeColumn instance' mapping."""
        return self._name_to_column

    @property
    def index_to_id(self):
        """Returns 'dataframe index - entity_id' mapping."""
        return self._index_to_id

    @property
    def new_col_names_map(self):
        """Returns 'dataframe index - event_date' mapping."""
        return self._new_col_names_map

    @property
    def index_to_date(self) -> Dict[int, int]:
        """Returns 'dataframe index - event_date' mapping."""
        return self._index_to_date

    @property
    def trx_file_record_collection(self) -> TransactionFileRecordCollection:
        """Returns trx_file_record_collection used for frame building."""
        return self._trx_file_record_collection

    def _set_included_operators(self, included_operators: List[AfeOperator],
                                dimension_column: str, quantity_column: str,
                                remove_operators_used_no_date: bool = False):
        self._included_operators = included_operators
        DataFrameBuilder._remove_operators_by_qty_and_dim(dimension_column, quantity_column,
                                                          self._included_operators)

        self._included_operator_set_one = list(
            set(OPERATOR_SET_ONE) & set(included_operators))
        self._included_operator_set_two = list(
            set(OPERATOR_SET_TWO) & set(included_operators))
        if remove_operators_used_no_date:
            DataFrameBuilder.__remove_operators_used_no_date(self._included_operator_set_two)

    def execute(self, num_of_threads: int, dimension_column: str, quantity_column: str,
                feature_generation_settings: FeatureGenerationSettings):
        """Generates dataframe by calculating AFE features and values for every entity"""

        trx_record_collection = self._trx_file_record_collection
        included_operators = feature_generation_settings.included_operators.copy()
        remove_operators_used_no_date = False
        if feature_generation_settings.date_column is None:
            remove_operators_used_no_date = True
        self._set_included_operators(included_operators, dimension_column, quantity_column,
                                     remove_operators_used_no_date)
        horizons = feature_generation_settings.horizon_list
        horizons.sort()
        histogram = self._trx_stats.get_histogram(dimension_column)
        nearest_exponent = self.__get_nearest_exponent(histogram.original_to_compressed, horizons)
        date_column = None
        date_column_index = -1
        if feature_generation_settings.date_column is not None:
            date_column = feature_generation_settings.date_column
            date_column_index = trx_record_collection.date_col_map[date_column.column_name]

        hash_column_map = self._add_columns_to_frame(horizons,
                                                     date_column,
                                                     feature_generation_settings.date_resolution,
                                                     feature_generation_settings.date_offset,
                                                     dimension_column,
                                                     quantity_column, nearest_exponent)

        splits = DataFrameBuilder._split(self._target_file_record_collection, num_of_threads)
        self._set_index_mappings(splits)

        new_entity_index_map = self._get_new_entity_index_map(splits, trx_record_collection.entity_index_map)

        # pylint: disable=c-extension-no-member
        data_frame_builder_helper.execute(
            self._frame.data_frame,
            num_of_threads,
            date_column_index,
            trx_record_collection.d_map[dimension_column],
            trx_record_collection.q_map[quantity_column],
            splits,
            horizons,
            feature_generation_settings.date_offset,
            feature_generation_settings.date_resolution.value,
            histogram.original_to_compressed,
            trx_record_collection.dates,
            trx_record_collection.q_arrays,
            trx_record_collection.d_arrays,
            new_entity_index_map,
            self._target_file_record_collection.event_dates,
            [a.value for a in self._included_operators],
            [a.value for a in self._included_operator_set_one],
            [a.value for a in self._included_operator_set_two],
            hash_column_map,
            nearest_exponent
        )
        if self._target_descriptor.target_column is not None:
            if self.target_array is None:
                self._generate_target()

    def _get_new_entity_index_map(self, splits, entity_index_map):
        new_entity_index_map = {}
        for i in range(splits.shape[0]):
            for j in range(splits.shape[1]):
                frame_index = i + j * splits.shape[0]
                record_index = splits[i][j]
                if record_index != -1:
                    entity = self._target_file_record_collection.entities[record_index]
                    if entity in entity_index_map:
                        new_entity_index_map[record_index] = entity_index_map[entity]
                    self._record_index_target_index_map[frame_index] = record_index
        return new_entity_index_map

    @staticmethod
    def _remove_operators_by_qty_and_dim(dimension_column: str, quantity_column: str,
                                         included_operators: List[AfeOperator]):
        if quantity_column != AfeStaticObjects.empty_quantity_column or \
                dimension_column == AfeStaticObjects.empty_dimension_column:
            if AfeOperator.Mode in included_operators:
                included_operators.remove(AfeOperator.Mode)
            if AfeOperator.CountDistinct in included_operators:
                included_operators.remove(AfeOperator.CountDistinct)

    @staticmethod
    def __remove_operators_used_no_date(included_operator_set_two):
        remove_included_operator_set_two = OPERATOR_SET_TWO
        for operator in remove_included_operator_set_two:
            if operator in included_operator_set_two:
                included_operator_set_two.remove(operator)

    def execute_for_afe_columns(self, num_of_threads: int, afe_columns: Dict[str, BaseAfeFeature]):
        """Builds dataframe by aggregating data according to given afe columns."""

        counter = 0

        splits = DataFrameBuilder._split(self._target_file_record_collection, num_of_threads)
        self._set_index_mappings(splits)

        column_index_name_map: Dict[int, str] = {}
        for date_col, afe_columns_for_date_col in self._group_features_by_date_column(afe_columns).items():
            if date_col is not None:
                self._trx_file_record_collection.sort(date_col)

            # entity_index_map should be regenerated after sorting trx_file entity index map
            new_entity_index_map = self._get_new_entity_index_map(splits,
                                                                  self._trx_file_record_collection.entity_index_map)

            for feature_name, feature in afe_columns_for_date_col.items():
                afe_column = feature.afe_column
                counter += 1
                dim_group_id = afe_column.group_id
                date_col_name = afe_column.date_column.column_name if afe_column.date_column is not None else ""
                LogHelper.info(f"Frame builder executing for afe column {counter}/{len(afe_columns)} : "
                               f"{afe_column.dimension_name}-{dim_group_id}-"
                               f"{afe_column.quantity_name}-{afe_column.operator}-"
                               f"{date_col_name}-{afe_column.time_window}")

                op_to_calculate = afe_column.operator

                included_operators = [op_to_calculate] if op_to_calculate in self.operators else []

                self._set_included_operators(included_operators, afe_column.dimension_name,
                                             afe_column.quantity_name)

                horizons = [afe_column.time_window]

                histogram = self._trx_stats.get_histogram(afe_column.dimension_name)
                nearest_exponent = self.__get_nearest_exponent(histogram.original_to_compressed, horizons)

                hash_val, column_index = self._add_column_to_frame_for_feature(feature, nearest_exponent)
                column_index_name_map[column_index] = feature_name

                date_column_index = -1
                if afe_column.date_column is not None:
                    date_column_index = \
                        self._trx_file_record_collection.date_col_map[afe_column.date_column.column_name]

                original_to_compressed = histogram.original_to_compressed.copy()
                original_to_compressed[np.int16(TransactionFileRecordReader.UNKNOWN_DIMENSION_VALUE)] = \
                    np.int16(TransactionFileRecordReader.UNKNOWN_DIMENSION_VALUE)

                # pylint: disable=c-extension-no-member
                data_frame_builder_helper.execute(
                    self._frame.data_frame,
                    num_of_threads,
                    date_column_index,
                    self._trx_file_record_collection.d_map[afe_column.dimension_name],
                    self._trx_file_record_collection.q_map[afe_column.quantity_name],
                    splits,
                    horizons,
                    afe_column.offset,
                    afe_column.date_resolution.value,
                    original_to_compressed,
                    self._trx_file_record_collection.dates,
                    self._trx_file_record_collection.q_arrays,
                    self._trx_file_record_collection.d_arrays,
                    new_entity_index_map,
                    self._target_file_record_collection.event_dates,
                    [a.value for a in self._included_operators],
                    [a.value for a in self._included_operator_set_one],
                    [a.value for a in self._included_operator_set_two],
                    {hash_val: column_index},
                    nearest_exponent,
                    dim_group_id if dim_group_id is not None else -1,
                    True,
                    afe_column.dimension_name == AfeStaticObjects.empty_dimension_column
                )
        self._update_column_names_in_frame(column_index_name_map)

    def _update_column_names_in_frame(self, column_index_name_map: Dict[int, str]):
        new_col_names_map = {}
        for col_index, feature_name in column_index_name_map.items():
            old_name = self._frame.get_column_name_by_index(col_index)
            new_col_names_map[old_name] = feature_name
        self._new_col_names_map = new_col_names_map
        self._frame.rename_columns(new_col_names_map)

    @staticmethod
    def _group_features_by_date_column(afe_columns: Dict[str, BaseAfeFeature]) -> Dict[str, Dict[str, BaseAfeFeature]]:
        features_grouped_by_date_col = {}

        for feature_name, feature in afe_columns.items():
            date_col_name = feature.afe_column.date_column.column_name \
                if feature.afe_column.date_column is not None else None
            if date_col_name in features_grouped_by_date_col:
                features_grouped_by_date_col[date_col_name][feature_name] = feature
            else:
                features_grouped_by_date_col[date_col_name] = {feature_name: feature}
        return features_grouped_by_date_col

    def get_frame_as_pandas_dataframe(self) -> pd.DataFrame:
        """Returns frame as pandas dataframe with additional entity id and event date columns"""
        data_frame = pd.DataFrame()

        record_indices = [self._record_index_target_index_map[i] for i in range(self._frame.data_frame.shape[0])]
        entities = [self._target_file_record_collection.entities[i] for i in record_indices]
        data_frame[AfeStaticObjects.distinct_entities_entity_column_name] = entities

        if self._target_descriptor.date_column is not None:
            dates = [date_helper.get_integer_as_date(self._target_file_record_collection.event_dates[i]) for i in
                     record_indices]
            data_frame[AfeStaticObjects.event_date_column_name] = dates

        for col in self._frame.column_name_map:
            data_frame[col] = self._frame.get_value(col)
        return data_frame

    def _set_index_mappings(self, splits: np.ndarray):
        self._index_to_id = {}
        self._index_to_date = {}
        for i in range(splits.shape[0]):
            for j in range(splits.shape[1]):
                index = i + j * splits.shape[0]
                record_index: int = splits[i][j]
                if record_index != -1:
                    entity_id = self._target_file_record_collection.entities[record_index]
                    date = self._target_file_record_collection.event_dates[record_index]
                    self._index_to_id[index] = entity_id
                    self._index_to_date[index] = date

    def _get_maximum_number_of_frame_columns(self):
        max_count = 0
        for date_col in self._dimensions_per_date_col:
            num_horizons = len(self._horizons_per_date_col[date_col])
            count = (len(OPERATOR_SET_ONE) + 1) * num_horizons + len(OPERATOR_SET_TWO)

            max_num_of_dimension_groups = 0
            for dimension_column in self._dimensions_per_date_col[date_col]:
                histogram = self._trx_stats.get_histogram(dimension_column)
                num_groups_for_dimension = len(histogram.compressed_reverse_index)
                if num_groups_for_dimension > max_num_of_dimension_groups:
                    max_num_of_dimension_groups = num_groups_for_dimension
            max_for_date_col = count * max_num_of_dimension_groups
            max_count = max(max_for_date_col, max_count)
        return int(max_count + 1)  # +1 for target column

    def _add_column_to_frame_for_feature(self, feature: BaseAfeFeature, nearest_exponent: int):
        hash_column_map = {}
        stats = self._trx_stats
        histogram = stats.get_histogram(feature.afe_column.dimension_name)
        date_col_name = self._get_date_col_name_of_afe_column(feature.afe_column)
        date_column = AfeDateColumn(column_name=date_col_name)
        afe_column = self._get_base_afe_column_instance(date_column, feature.afe_column.dimension_name,
                                                        feature.afe_column.quantity_name,
                                                        feature.afe_column.date_resolution, feature.afe_column.offset)

        operator = feature.afe_column.operator
        horizon = feature.afe_column.time_window
        allowed_dim_values = None if feature.afe_column.group_id is None else [feature.afe_column.group_id]
        if operator in self._included_operator_set_one:
            afe_column.operator = operator
            afe_column.time_window = horizon
            self._add_columns_to_frame_with_indices(afe_column, histogram, nearest_exponent, hash_column_map,
                                                    allowed_dim_values)
        elif operator in self._included_operator_set_two:
            afe_column.operator = operator
            afe_column.time_window = horizon
            self._add_columns_to_frame_with_indices(afe_column, histogram, nearest_exponent, hash_column_map,
                                                    allowed_dim_values)
        elif AfeOperator.Ratio in self._included_operators:
            afe_column.operator = AfeOperator.Ratio
            afe_column.time_window = horizon
            self._add_columns_to_frame_with_indices(afe_column, histogram, nearest_exponent, hash_column_map,
                                                    allowed_dim_values)

        return next(iter(hash_column_map.items()))

    def _add_columns_to_frame(self, horizons: List[int],
                              date_column: AfeDateColumn, resolution: DateResolution, offset: int,
                              dimension_name: str, quantity_name: str,
                              nearest_exponent: int,
                              allowed_dim_values: List[int] = None):
        # pylint: disable=too-many-arguments
        hash_column_map = {}
        max_horizon = max(horizons)
        stats = self._trx_stats
        histogram = stats.get_histogram(dimension_name)
        afe_column = self._get_base_afe_column_instance(date_column, dimension_name, quantity_name, resolution, offset)
        for operator in self._included_operator_set_one:
            for horizon in horizons:
                afe_column.operator = operator
                afe_column.time_window = horizon
                self._add_columns_to_frame_with_indices(afe_column, histogram, nearest_exponent, hash_column_map,
                                                        allowed_dim_values)
        for operator in self._included_operator_set_two:
            afe_column.operator = operator
            afe_column.time_window = max_horizon
            self._add_columns_to_frame_with_indices(afe_column, histogram, nearest_exponent, hash_column_map,
                                                    allowed_dim_values)
        if AfeOperator.Ratio in self._included_operators:
            afe_column.operator = AfeOperator.Ratio
            for horizon in horizons:
                afe_column.time_window = horizon
                self._add_columns_to_frame_with_indices(afe_column, histogram, nearest_exponent, hash_column_map,
                                                        allowed_dim_values)

        return hash_column_map

    @classmethod
    def _get_base_afe_column_instance(cls, date_column: AfeDateColumn, dimension_name: str,
                                      quantity_name: str, resolution: DateResolution, offset: int) -> AfeColumn:
        afe_column = AfeColumn()
        afe_column.date_column = date_column
        afe_column.dimension_name = dimension_name
        afe_column.quantity_name = quantity_name
        afe_column.date_resolution = resolution
        afe_column.offset = offset
        return afe_column

    def _add_columns_to_frame_with_indices(self, base_afe_column: AfeColumn, histogram: Histogram16,
                                           nearest_exponent: int,
                                           hash_column_map: Dict[int, int],
                                           allowed_dim_values: List[int] = None):

        entire_index_list = list(histogram.index.values())
        for key, value in histogram.compressed_reverse_index.items():
            afe_column = copy.deepcopy(base_afe_column)
            group_id = key
            if allowed_dim_values and group_id not in allowed_dim_values:
                continue
            list_of_values = value
            dim_values = []
            for item in list_of_values:
                dim_values.append(histogram.index[item])

            complement = [x for x in entire_index_list if x not in dim_values]
            in_out = True
            if complement:
                in_out = len(dim_values) < len(complement)

            afe_column.group_id = group_id
            afe_column.set = dim_values if in_out else complement
            afe_column.in_out = in_out
            afe_column.set_column_name()

            column_name = afe_column.column_name
            group_name = afe_column.get_group_name()

            self._name_to_column[column_name] = afe_column
            if group_name not in self._group_to_name:
                self._group_to_name[group_name] = []
            self._group_to_name[group_name].append(column_name)

            self._frame.try_add(column_name)
            col_index = self._frame.column_name_map[column_name]

            self._add_column_to_map(hash_column_map, afe_column, col_index, nearest_exponent)

    @staticmethod
    def _split(collection: TargetFileRecordCollection, num_of_threads: int) -> \
            np.ndarray:
        _dict = np.full((num_of_threads, math.ceil(collection.sampled_count / num_of_threads)), -1, dtype=np.int32)

        if collection.sampled:
            indices = collection.indices
            for i, index in enumerate(indices):
                _dict[i % num_of_threads][math.floor(i / num_of_threads)] = index
        else:
            for i in range(collection.sampled_count):
                _dict[i % num_of_threads][math.floor(i / num_of_threads)] = i
        return _dict

    def clean_frame(self):
        """removes all columns from frame which are not in 'stayers'(except 'target_column')"""
        for column_name in self._frame.get_column_names():
            self._frame.remove(column_name)

    def _generate_target(self):
        record_count = self._target_file_record_collection.sampled_count
        base_target_column = self._target_descriptor.target_column
        target_column_type = base_target_column.target_column_type
        if target_column_type == AfeTargetColumnType.Binary:
            target_array = []
            positive_category = base_target_column.binary_target_info.positive_category
            negative_category = base_target_column.binary_target_info.negative_category
            if self._target_file_record_collection.sampled:
                indices: np.array = self._target_file_record_collection.indices
                for index in indices:
                    target_binary = self._target_file_record_collection.target_binary_values[index]
                    target_val = positive_category if target_binary == BinaryTargetClass.POSITIVE.value \
                        else negative_category
                    target_array.append(target_val)
            else:
                for i in range(record_count):
                    target_binary = self._target_file_record_collection.target_binary_values[i]
                    target_val = positive_category if target_binary == BinaryTargetClass.POSITIVE.value \
                        else negative_category
                    target_array.append(target_val)

            self.target_array = np.array(target_array)
        elif target_column_type == AfeTargetColumnType.Scalar:
            self.__generate_target_for_multi_class_and_scalar(
                record_count, self._target_file_record_collection.target_scalar_values)
        elif target_column_type == AfeTargetColumnType.MultiClass:
            self.__generate_target_for_multi_class_and_scalar(
                record_count, self._target_file_record_collection.target_multi_class_values)
        else:
            raise KnownException("Unknown target type :" + str(target_column_type))

    def __generate_target_for_multi_class_and_scalar(self, record_count: int, target_values):
        target_array = []
        if self._target_file_record_collection.sampled:
            indices = self._target_file_record_collection.indices
            for index in indices:
                target_array.append(target_values[index])
        else:
            for i in range(record_count):
                target_array.append(target_values[i])
        self.target_array = np.array(target_array)

    @staticmethod
    def __get_nearest_exponent(index_map: Dict, horizons):
        max_index_map_val = max(index_map.values())
        max_afe_operator_val = max([a.value for a in AfeOperator])
        max_val = max([max_index_map_val, max_afe_operator_val, max(horizons)])
        nearest_exponent = 10 ** (math.floor(math.log10(max_val)) + 1)
        return nearest_exponent

    @staticmethod
    def _add_column_to_map(hash_column_map, base_afe_column: AfeColumn, col_index, nearest_exponent: int):
        val = (nearest_exponent ** 2 * base_afe_column.group_id) + \
              (nearest_exponent ** 1 * base_afe_column.operator.value) + \
              base_afe_column.time_window
        hash_column_map[val] = col_index
        return col_index

    @staticmethod
    def _get_date_col_name_of_afe_column(afe_col: AfeColumn):
        return afe_col.date_column.column_name if afe_col.date_column is not None else ""



"""Includes helper functions for enum classes."""
from typing import Optional

from organon.fl.core.enums.column_measurement_type import ColumnMeasurementType
from organon.fl.core.enums.column_native_type import ColumnNativeType


def convert_measurement_to_native_type(measurement_type: ColumnMeasurementType) -> Optional[ColumnNativeType]:
    """Converts ColumnMeasurementType to ColumnNativeType"""
    if measurement_type == ColumnMeasurementType.NOMINAL:
        return ColumnNativeType.String
    if measurement_type == ColumnMeasurementType.NUMERIC:
        return ColumnNativeType.Numeric
    if measurement_type == ColumnMeasurementType.DATE:
        return ColumnNativeType.Date
    if measurement_type == ColumnMeasurementType.OTHER:
        return ColumnNativeType.Other
    return None



"""Includes SearchType enum class."""
from enum import Enum, auto


class SearchType(Enum):
    """Search types for hyperparameter optimization"""
    GRID = auto()
    RANDOM = auto()
    MODEL_BASED = auto()



"""todo"""
from typing import Optional

from organon.fl.core.businessobjects.dataframe import DataFrame
from organon.idq.domain.businessobjects.statistics.data_source_statistics import DataSourceStatistics
from organon.idq.domain.businessobjects.statistics.population_statistics import PopulationStatistics
from organon.idq.domain.businessobjects.statistics.sample_statistics import SampleStatistics


class DqCalculationResult:
    """todo"""

    def __init__(self):
        self.calculation_name: str = None
        self.data_source_stats: Optional[DataSourceStatistics] = None
        self.sample_stats: Optional[SampleStatistics] = None
        self.population_stats: Optional[PopulationStatistics] = None
        self.sample_data: Optional[DataFrame] = None
        self.column_metadata_list = None



"""Includes PriorityQueue"""


class PriorityQueue:
    """Priority queue implementation using a list"""
    def __init__(self):
        self.queue = []

    def is_empty(self):
        """Returns true if queue is empty"""
        return len(self.queue) == 0

    def push(self, item, priority):
        """
        item already in priority queue with smaller priority:
        -> update its priority
        item already in priority queue with higher priority:
        -> do nothing
        if item not in priority queue:
        -> push it
        """
        for index, (i, i_priority) in enumerate(self.queue):
            if i == item:
                if i_priority >= priority:
                    break
                del self.queue[index]
                self.queue.append((item, priority))
                break
        else:
            self.queue.append((item, priority))

    def pop(self):
        """return item with highest priority value and remove it from queue"""
        if self.is_empty():
            raise ValueError("Queue is empty")
        max_idx = 0
        for index, (_, i_priority) in enumerate(self.queue):
            if self.queue[max_idx][1] < i_priority:
                max_idx = index
        item, priority = self.queue[max_idx]
        del self.queue[max_idx]
        return item, priority



"""This module includes PredictionResult"""
from typing import List


class PredictionResult:
    """Class for PredictionResult"""

    def __init__(self, actual: list):
        self.predictions: List[float] = None
        self.actual: List[float] = actual
        self.mean_square_err: float = None



"""
This module includes TransactionFileStats class.
"""
from typing import Dict

import numpy as np

from organon.fl.core.businessobjects.histogram_16 import Histogram16


class TransactionFileStats:
    """
    Statistic details for TransactionFile
    """

    def __init__(self):
        self.__histograms: Dict[str, Histogram16[str]] = {}  # pylint: disable=unsubscriptable-object

    def get_histogram(self, dimension_column_name: str) -> Histogram16:
        """Returns histogram corresponding to given dimension column"""
        return self.__histograms[dimension_column_name]

    def build_indices(self, compression_ratio: float, max_cardinality: int):
        """build indices for each histogram"""
        for histogram in self.__histograms.values():
            histogram.build_indices_with_params(compression_ratio, max_cardinality)

    def add(self, dimension_column_name: str):
        """Adds a new histogram corresponding to given dimension column"""
        if dimension_column_name not in self.__histograms:
            self.__histograms[dimension_column_name] = Histogram16()

    def increment(self, dimension_column_name: str, column_value: str, frequency: np.int64 = None):
        """add a value to column data"""
        self.__histograms[dimension_column_name].add(column_value, frequency)

    def get_index(self, column: str, column_value: str) -> np.short:
        """return index for a column value"""
        return self.__histograms[column].get_index_value(column_value)



"""Includes SamplingOutput class."""
import pandas as pd


class SamplingOutput:
    """Output of Sampling service"""

    def __init__(self):
        self.train_data: pd.DataFrame = None
        self.test_data: pd.DataFrame = None



"""Includes BaseComparisonInput class."""
from typing import List

from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType


class BaseComparisonInput:
    """Base input dto for time series ad traffic light comparisons"""

    def __init__(self):
        self.data_entity: DataEntityType = None
        self.data_entity_name: str = None
        self.test_group: DqTestGroupType = None
        self.past_series: List[float] = None
        self.current_value: float = None



"""
This module keeps DateInterval class.
"""
from datetime import datetime
from typing import Callable, Union

from organon.fl.core.enums.date_interval_type import DateIntervalType


class DateInterval:
    """
    This class is used for checking the date interval types.
    """

    def __init_subclass__(cls, **kwargs):
        raise TypeError("type 'DateInterval' is not an acceptable base type")

    def __init__(self, lower_bound: Union[datetime, int], upper_bound: Union[datetime, int],
                 interval_type: DateIntervalType):
        self.__interval_type: DateIntervalType = interval_type
        self.__lower_bound: Union[datetime, int] = lower_bound
        self.__upper_bound: Union[datetime, int] = upper_bound
        self.__func: Callable[[Union[datetime, int]], bool] = self.get_func(self.__interval_type)

    @property
    def interval_type(self) -> DateIntervalType:
        """
        Returns the date interval type.
        :return: Date interval type of the current object
        """
        return self.__interval_type

    @property
    def lower_bound(self) -> Union[datetime, int]:
        """
        Returns the lower bound datetime.
        :return: lower bound datetime object
        """
        return self.__lower_bound

    @property
    def upper_bound(self) -> Union[datetime, int]:
        """
        Returns the upper bound datetime.
        :return: upper bound datetime object
        """
        return self.__upper_bound

    @property
    def func(self) -> Callable[[Union[datetime, int]], bool]:
        """
        Returns a bool function which takes a datetime object or timesstamp value as a parameter.
        :return: bool function which takes a datetime object or timestamp value as a parameter
        """
        return self.__func

    def get_func(self, interval_type: DateIntervalType) -> Callable[[Union[datetime, int]], bool]:
        """
        Returns a function according to the date interval type.
        :param interval_type: date interval type
        :return: bool function which takes a datetime object as a parameter
        """

        if interval_type == DateIntervalType.CLOSED_CLOSED:
            return self.__closed_closed
        if interval_type == DateIntervalType.OPEN_CLOSED:
            return self.__open_closed
        if interval_type == DateIntervalType.CLOSED_OPEN:
            return self.__closed_open
        if interval_type == DateIntervalType.OPEN_OPEN:
            return self.__open_open
        return self.__default

    def __closed_closed(self, date: Union[datetime, int]) -> bool:
        """
        Checks if the datetime object has closed closed relation between the intervals.
        :param date: datetime object or timestamp value
        :return: bool object which checks whether the datetime object has closed closed relation between the
        intervals
        """
        return self.__lower_bound <= date <= self.__upper_bound

    def __open_closed(self, date: Union[datetime, int]) -> bool:
        """
        Checks if the datetime object has open closed relation between the intervals.
        :param date: datetime object or timestamp
        :return: bool object which checks whether the datetime object has open closed relation between the intervals
        """
        return self.__lower_bound < date <= self.__upper_bound

    def __closed_open(self, date: Union[datetime, int]) -> bool:
        """
        Checks if the datetime object has closed open relation between the intervals.
        :param date: datetime object
        :return: bool object which checks whether the datetime object has closed open relation between the intervals
        """
        return self.__lower_bound <= date < self.__upper_bound

    def __open_open(self, date: Union[datetime, int]) -> bool:
        """
        Checks if the datetime object has open open relation between the intervals.
        :param date: datetime object
        :return: bool object which checks whether the datetime object has open open relation between the intervals
        """
        return self.__lower_bound < date < self.__upper_bound

    def __default(self, date: Union[datetime, int]) -> bool:
        """
        Checks if the datetime object has default relation between the intervals.
        :param date: datetime object
        :return: bool object which checks whether the datetime object has default relation between the intervals
        """
        return self.__lower_bound <= date <= self.__upper_bound

    def contains(self, val: Union[datetime, int]) -> bool:
        """
        Checks if the function contains datetime object.
        :param val: datetime object
        :return: bool object which checks whether the function contains datetime object.
        """
        return self.__func(val)



"""Module for CovarianceGenerator"""
from typing import List

import numpy as np
# pylint: disable=no-name-in-module
from organon.fl.mathematics.sweep.pyx_files.sweep_helper import compute_covariances_c

from organon.fl.mathematics.linearalgebra.IBlas import IBlas
from organon.fl.mathematics.sweep.covariance_info import CovarianceInfo
from organon.fl.mathematics.sweep.data_classes.transformation import Transformation
from organon.fl.mathematics.sweep.matrices.csc_matrix import CscMatrix
from organon.fl.mathematics.sweep.matrices.dense_matrix import DenseMatrix
from organon.fl.mathematics.sweep.matrices.i_matrix import IMatrix


class CovarianceGenerator:
    """Class for JIT CovarianceGenerator. Firstly calculates diagonals of covariance and correlation
    matrices. After that calculates the corresponding columns and rows of the matrices."""

    @property
    def covariance_info(self):
        """Get the covariance info"""
        return self._covariance_info

    @property
    def biased(self):
        """Get the bias info"""
        return self._biased

    def __init__(self, biased: bool, matrix: IMatrix, transformations: List[Transformation], weight_column: np.ndarray,
                 blas_service: IBlas):
        self.validate(matrix, transformations, weight_column)

        self._biased = biased
        self._matrix = matrix
        if matrix.data.dtype.type == np.float32:
            self.dot = blas_service.sdot
        else:
            self.dot = blas_service.ddot

        self._variable_count = self._matrix.shape[1]
        # self._indices = indices
        self._transformations = transformations
        self._weight_column = weight_column

        self._is_computed = np.zeros((self._variable_count,), dtype=np.uint8)

        self._covariance_info: CovarianceInfo = None

    @staticmethod
    def validate(matrix: IMatrix, transformations: List[Transformation], weight_column: np.ndarray):
        """Validate the matrix, transformations and weight column"""
        if matrix is None:
            raise ValueError("Matrix can not be None as input to JitCovarianceMatrixGenerator")
        if transformations is None:
            raise ValueError("Transformations can not be None as input to JitCovarianceMatrixGenerator")
        if weight_column is None:
            raise ValueError("Weight_column can not be None as input to JitCovarianceMatrixGenerator")
        if weight_column.shape[0] != matrix.shape[1]:
            raise ValueError(f"Matrix shape and Weight Column shape don't fit. "
                             f"weight_column.shape = {weight_column.shape}, matrix.shape = {matrix.shape}")

    def initialize(self):
        """initialize the covariance info using covariance matrix and dimension"""
        dim = self._variable_count

        self._matrix.data[:] = CovarianceGenerator.get_transformed_array(self._matrix.data, self._transformations[0])

        matrix = np.empty(shape=(dim, dim), dtype=self._matrix.data.dtype)
        matrix.fill(float('nan'))
        covariance_matrix = matrix
        self._covariance_info = CovarianceInfo(covariance_matrix, dim)

    @staticmethod
    def get_transformed_array(inp, transformation: Transformation) -> np.ndarray:
        """Apply transformation to input"""
        return np.array([transformation.evaluate(item) for item in inp])

    def compute_variances(self):
        """Calculates diagonals of correlation and covariance matrix"""
        sigma = self._covariance_info.covariance_matrix
        corr_matrix = self._covariance_info.correlation_matrix
        for i in range(self._variable_count):
            if ~np.isnan(sigma[i, i]):
                continue
            col_i = self._matrix.get_column(i)
            variance = self.dot(self._matrix.shape[0], col_i, 1, col_i, 1)
            variance /= self._matrix.shape[0] - 1

            sigma[i, i] = variance
            corr_matrix[i, i] = 1.0

    def compute_covariances_c(self, variable: int, n_thread: int, bin_size: int):
        """
        Compute Covariance using Cython
        """
        compute_covariances_c(self._matrix,
                              self._matrix.means,
                              self._covariance_info.covariance_matrix,
                              self._covariance_info.correlation_matrix,
                              self._is_computed,
                              variable,
                              bin_size,
                              n_thread)

    def compute_covariances(self, variable: int):
        """Computes covariances of @variable and other elements
        This method computes covariances JIT."""
        sigma = self._covariance_info.covariance_matrix
        corr_matrix = self._covariance_info.correlation_matrix

        col_variable = self._matrix.get_column(variable)

        if isinstance(self._matrix, DenseMatrix):
            covariances = self._matrix.mul_vector(col_variable)
        elif isinstance(self._matrix, CscMatrix):
            dense_m = self._matrix.to_dense()
            dense_m -= np.mean(dense_m, axis=0)
            covariances = np.dot(dense_m.T, col_variable)
        else:
            raise NotImplementedError
        covariances /= self._matrix.shape[0] - 1
        i = variable

        for j in range(self._variable_count):
            if self._is_computed[j]:
                continue
            if ~np.isnan(sigma[i, j]):
                continue

            sigma[i, j] = sigma[j, i] = covariances[j]

            nan_i = np.isnan(sigma[i, i])
            zero_i = sigma[i, i] <= 0

            if i == j:
                corr_matrix[i, i] = float('nan') if nan_i else 1.0
            else:
                nan_j = np.isnan(sigma[j, j])
                zero_j = sigma[j, j] <= 0

                if np.isnan(sigma[i, j]) or nan_i or nan_j:
                    corr_matrix[i, j] = float('nan')

                elif zero_i or zero_j:
                    corr_matrix[i, j] = 0.0

                else:
                    corr_matrix[i, j] = sigma[i, j] / np.sqrt(sigma[i, i] * sigma[j, j])

                corr_matrix[j, i] = corr_matrix[i, j]

    def add(self, variable):
        """Adds a variable to calculated variables"""
        self._is_computed[variable] = True

    def get_density(self):
        """Returns density(not empty elements) of the covariance matrix"""
        sigma = self._covariance_info.covariance_matrix
        return np.count_nonzero(~np.isnan(sigma)) / sigma.size



"""Includes DqControlType enum definition."""
from enum import Enum, auto


class DqControlType(Enum):
    """Control types in DQ."""
    EMPTY_TABLE = auto()
    TABLE_COLUMNS = auto()
    STABLE_COLUMN = auto()
    UNEXPECTED_NUMERICAL_VALUES = auto()
    UNEXPECTED_NOMINAL_VALUES = auto()
    DUPLICATE_KEYS = auto()
    COLUMN_MEAN = auto()
    DATA_SOURCE_STATS_TL_CONTROL = auto()
    DATA_SOURCE_STATS_TIME_SERIES_CONTROL = auto()
    PSI_NOMINAL_VALUES = auto()
    PSI_NUMERIC_VALUES = auto()



"""Includes HPOptimizationSettings class"""
from dataclasses import dataclass
from typing import List, Dict, Optional

import pandas as pd

from organon.ml.modelling.algorithms.core.enums.modeller import Modeller
from organon.ml.modelling.model_selection.settings.enums.search_type import SearchType


@dataclass
class HPOptimizationSettings:
    """Settings for HPOService"""
    train_data: pd.DataFrame
    target_data: pd.Series
    search_method: SearchType
    modellers: List[Modeller] = None
    modeller_params: List[Dict[str, list]] = None
    cv_fold: int = None
    scoring_metrics: Optional[List[str]] = None
    test_data: pd.DataFrame = None
    test_target_data: pd.Series = None
    search_params: dict = None



"""Includes BaseObjectClassificationService"""
import abc
import math
import os
from typing import Dict, List, Optional, Tuple, Union, TYPE_CHECKING

import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score, f1_score, recall_score, precision_score
from sklearn.utils import class_weight

from organon.fl.logging.helpers.log_helper import LogHelper
from organon.ml.common.enums.classification_type import ClassificationType
from organon.ml.common.enums.color_type import ColorType
from organon.ml.common.enums.optimizer_type import OptimizerType
from organon.ml.object_classification.common.helpers import throw_val_ex_with_log
from organon.ml.object_classification.domain.common.helpers import get_data_source_settings
from organon.ml.object_classification.domain.common.object_clf_settings_validator import ObjectClfSettingsValidator
from organon.ml.object_classification.common.object_classification_constants import ObjectClassificationConstants
from organon.ml.object_classification.domain.data_sources.base_data_source_handler import BaseDataSourceHandler
from organon.ml.object_classification.domain.data_sources.data_source_handler_factory import DataSourceHandlerFactory
from organon.ml.object_classification.domain.data_sources.tf_dataset_data_source_handler import \
    TfDatasetDataSourceHandler
from organon.ml.object_classification.domain.data_sources.unlabeled_dir_data_source_handler import \
    UnlabeledDirDataSourceHandler
from organon.ml.object_classification.domain.objects.object_clf_settings import \
    ObjectClfSettings
from organon.ml.object_classification.domain.common.base_model_generator import BaseModelGenerator

if TYPE_CHECKING:
    import tensorflow as tf  # noqa


class BaseObjectClassificationService(metaclass=abc.ABCMeta):
    """Base Object classification service"""

    def __init__(self, settings: ObjectClfSettings):
        self._settings = settings
        self.model: "tf.keras.Model" = None

    def fit(self):
        """Fit classifier"""

        ObjectClfSettingsValidator.check_settings(self._settings)

        class_names = os.listdir(self._settings.train_data_dir)

        LogHelper.info("Reading train data..")
        train_data, validation_data = self._generate_image_data()

        class_weights = self._check_class_weights(class_names, train_data)

        LogHelper.info("Preparing model..")
        base_model, inputs = BaseModelGenerator.get_base_model(self._settings.image_size,
                                                               self._settings.color_mode,
                                                               self._settings.pretrained_model_settings,
                                                               self._settings.data_augmentation_settings,
                                                               self._settings.random_seed)

        self.model = self._create_final_dense_layer(base_model, inputs, layer_num=len(class_names))

        LogHelper.info("Model training started..")
        self._train_model(train_data, validation_data, class_weights, base_model)

        if self._settings.fine_tuning_settings is not None:
            LogHelper.info("Fine tuning started..")
            self._train_model(train_data, validation_data, class_weights, base_model, fine_tuning=True)

    def evaluate(self, data: Union["tf.data.Dataset", str], average: str = None, return_df: bool = True,
                 is_processed: bool = False) -> Tuple[Dict[str, float], Optional[pd.DataFrame]]:
        """Predict and evaluate predictions"""

        data_source = DataSourceHandlerFactory.get_data_source_handler(data, self._settings)
        if isinstance(data_source, UnlabeledDirDataSourceHandler):
            msg = "Unlabeled sources can't be evaluated!"
            LogHelper.error(msg)
            raise TypeError(msg)

        probs_df = self.predict(data, is_processed)

        y_true = probs_df.loc[:, ObjectClassificationConstants.LABELS_COL_NAME]
        y_pred = pd.to_numeric(probs_df.loc[:, ObjectClassificationConstants.PREDICTON_COL_NAME])

        average = self._get_avg_parameter(average)

        metrics = {"accuracy": accuracy_score(y_true, y_pred),
                   "recall": recall_score(y_true, y_pred, average=average),
                   "f1": f1_score(y_true, y_pred, average=average),
                   "precision": precision_score(y_true, y_pred, average=average)}
        return (metrics, probs_df) if return_df else (metrics, None)

    def predict(self, data: Union["tf.data.Dataset", str], is_processed: bool = False) -> pd.DataFrame:
        """Predict object in image"""

        data_source = DataSourceHandlerFactory.get_data_source_handler(data, self._settings)

        probs = self._get_probs(data_source, is_processed)

        predictions = self._get_predictions(probs)

        pred_df = data_source.initialize_probs_df()
        pred_df[ObjectClassificationConstants.PREDICTON_COL_NAME] = predictions

        return pred_df

    def predict_proba(self, data: Union["tf.data.Dataset", str], is_processed: bool = False) -> pd.DataFrame:
        """Predict probabilities of object in image"""

        data_source = DataSourceHandlerFactory.get_data_source_handler(data, self._settings)

        probs = self._get_probs(data_source, is_processed)

        probs_df: pd.DataFrame = data_source.initialize_probs_df()

        probs_df = self._set_probas(probs_df, probs)

        return probs_df

    def get_processed_data(self, data: Union["tf.data.Dataset", str]) -> "tf.data.Dataset":
        """Convert data into a format the model can predict"""
        data_source = DataSourceHandlerFactory.get_data_source_handler(data, self._settings)
        data_source.process_data()
        return data_source.dataset

    @abc.abstractmethod
    def _set_probas(self, probs_df: pd.DataFrame, probs: np.ndarray) -> pd.DataFrame:
        raise NotImplementedError

    @abc.abstractmethod
    def _create_final_dense_layer(self, base_model: "tf.Tensor", inputs: "tf.keras.Input",
                                  layer_num: int) -> "tf.keras.Model":
        raise NotImplementedError

    @abc.abstractmethod
    def _get_predictions(self, probs: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_avg_parameter(self, average: Optional[str]) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_compiled_model(self, fine_tuning: bool = False) -> "tf.keras.Model":
        raise NotImplementedError

    @abc.abstractmethod
    def _get_output_layer(self, base_model: "tf.Tensor", layer_num: int):
        raise NotImplementedError

    @abc.abstractmethod
    def _generate_image_data(self):
        raise NotImplementedError

    def _get_probs(self, data: BaseDataSourceHandler, is_processed: bool = False) -> np.ndarray:

        if not is_processed:
            LogHelper.info("Data is not preprocessed! Processing data...")
            data.process_data()

        return self.model.predict(data.dataset)

    def _get_color_parameter(self) -> str:
        if self._settings.color_mode == ColorType.RGB:
            color_mode = "rgb"
        elif self._settings.color_mode == ColorType.RGBA:
            color_mode = "rgba"
        elif self._settings.color_mode == ColorType.GRAY:
            color_mode = "grayscale"
        else:
            return throw_val_ex_with_log("Color type is ambiguous!")

        return color_mode

    def _get_optimizer(self, fine_tuning: bool = False) -> "tf.keras.optimizers":
        import tensorflow as tf  # pylint: disable=import-outside-toplevel
        opt = self._settings.fine_tuning_settings.optimizer if fine_tuning \
            else self._settings.transfer_learning_settings.optimizer
        learning_rate = self._get_learning_rate(opt, fine_tuning)
        if opt == OptimizerType.ADAM:
            optimizer = tf.keras.optimizers.Adam(learning_rate)
        elif opt == OptimizerType.ADADELTA:
            optimizer = tf.keras.optimizers.Adadelta(learning_rate)
        elif opt == OptimizerType.ADAGRAD:
            optimizer = tf.keras.optimizers.Adagrad(learning_rate)
        elif opt == OptimizerType.ADAMAX:
            optimizer = tf.keras.optimizers.Adamax(learning_rate)
        elif opt == OptimizerType.FTRL:
            optimizer = tf.keras.optimizers.Ftrl(learning_rate)
        elif opt == OptimizerType.NADAM:
            optimizer = tf.keras.optimizers.Nadam(learning_rate)
        elif opt == OptimizerType.SGD:
            optimizer = tf.keras.optimizers.SGD(learning_rate)
        elif opt == OptimizerType.RMSPROP:
            optimizer = tf.keras.optimizers.RMSprop(learning_rate)
        else:
            return throw_val_ex_with_log("Optimizer is ambiguous!")

        return optimizer

    def _get_learning_rate(self, opt: OptimizerType, fine_tuning: bool = False) -> float:
        learning_rate = self._settings.fine_tuning_settings.learning_rate if fine_tuning \
            else self._settings.transfer_learning_settings.learning_rate

        if opt == OptimizerType.SGD:
            if not fine_tuning and self._settings.transfer_learning_settings.learning_rate == 0.001:
                learning_rate = 0.01
        return learning_rate

    def _get_steps_parameter(self, data: "tf.data.Dataset") -> int:
        return math.ceil(int(data.cardinality().numpy() // self._settings.batch_size)) - 1

    def _train_model(self, train_data: TfDatasetDataSourceHandler, validation_data: TfDatasetDataSourceHandler,
                     class_weights: Optional[Dict[int, float]], base_model: "tf.Tensor", fine_tuning: bool = False):
        import tensorflow as tf  # pylint: disable=import-outside-toplevel
        callbacks = [
            tf.keras.callbacks.EarlyStopping(
                patience=self._settings.fine_tuning_settings.early_stopping_patience if fine_tuning
                else self._settings.transfer_learning_settings.early_stopping),
            tf.keras.callbacks.TensorBoard()
        ]

        if fine_tuning:
            base_model.trainable = True

        self._get_compiled_model(fine_tuning)

        epochs = self._settings.fine_tuning_settings.epoch if fine_tuning \
            else self._settings.transfer_learning_settings.epoch

        self.model.fit(
            x=train_data.data,
            epochs=epochs,
            steps_per_epoch=self._get_steps_parameter(train_data.data),
            validation_data=validation_data.data,
            validation_steps=self._get_steps_parameter(validation_data.data),
            class_weight=class_weights,
            callbacks=callbacks)

    def _add_final_base_layers(self, base_model: "tf.Tensor") -> "tf.Tensor":
        import tensorflow as tf  # pylint: disable=import-outside-toplevel
        base_model = tf.keras.layers.GlobalAveragePooling2D()(base_model)
        base_model = tf.keras.layers.Dropout(self._settings.transfer_learning_settings.dropout)(base_model)
        return base_model

    def _generate_train_validation_data(self, label_mode: str) -> Tuple[
        TfDatasetDataSourceHandler, TfDatasetDataSourceHandler]:
        import tensorflow as tf  # pylint: disable=import-outside-toplevel
        if self._settings.validation_data_dir is not None:
            train_data = tf.keras.utils.image_dataset_from_directory(directory=self._settings.train_data_dir,
                                                                     batch_size=self._settings.batch_size,
                                                                     image_size=self._settings.image_size,
                                                                     shuffle=True,
                                                                     seed=self._settings.random_seed,
                                                                     color_mode=self._get_color_parameter(),
                                                                     label_mode=label_mode)

            validation_data = tf.keras.utils.image_dataset_from_directory(
                directory=self._settings.validation_data_dir,
                batch_size=self._settings.batch_size,
                image_size=self._settings.image_size,
                shuffle=False, seed=self._settings.random_seed,
                color_mode=self._get_color_parameter(),
                label_mode=label_mode)

        else:
            train_data, validation_data = tf.keras.utils.image_dataset_from_directory(
                directory=self._settings.train_data_dir,
                batch_size=self._settings.batch_size,
                image_size=self._settings.image_size,
                shuffle=True, seed=self._settings.random_seed,
                validation_split=self._settings.validation_data_ratio,
                subset="both", color_mode=self._get_color_parameter(), label_mode=label_mode)

        settings = get_data_source_settings(self._settings)

        return TfDatasetDataSourceHandler(train_data, settings), TfDatasetDataSourceHandler(validation_data, settings)

    def _check_class_weights(self, class_names: List[str], train_data: TfDatasetDataSourceHandler) -> Optional[
        Dict[int, float]]:

        class_dirs = [os.path.join(self._settings.train_data_dir, class_name) for class_name in class_names]

        n_of_samples = [len(os.listdir(directory)) for directory in class_dirs]

        n_of_total_samples = sum(n_of_samples)
        if n_of_total_samples < self._settings.batch_size:
            throw_val_ex_with_log(
                f"Batch size ({self._settings.batch_size}) can't be bigger than "
                f"total sample number ({n_of_total_samples})")

        labels = np.concatenate(train_data.get_labels(), axis=0)
        labels = labels.flatten() if self._settings.clf_mode == ClassificationType.BINARY else np.argmax(labels, axis=1)

        for (label, class_name, sample_number) in zip(labels, class_names, n_of_samples):
            LogHelper.info(f"Class {label}: {class_name}, {sample_number} samples")

        if len(set(n_of_samples)) != 1:
            LogHelper.info("Dataset is inbalanced")
            class_weights = class_weight.compute_class_weight(class_weight='balanced',
                                                              classes=np.unique(labels),
                                                              y=labels)
            return dict(enumerate(class_weights))
        return None



"""Includes StackingEnsembleRegressor class"""
from typing import Dict, Any

from sklearn.ensemble import StackingRegressor

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_regressor import BaseSklearnRegressor


class StackingEnsembleRegressor(BaseSklearnRegressor):
    """Wrapper for sklearn StackingRegressor"""

    @property
    def regressor_class(self):
        return StackingRegressor

    def _get_params_with_defaults(self) -> Dict[str, Any]:
        return StackingRegressor(estimators=[]).get_params()



"""Includes SingleSourceOutputReport class."""
from organon.idq.domain.reporting.objects.base_dq_output_report import BaseDqOutputReport


class SingleSourceOutputReport(BaseDqOutputReport):
    """Output report for DQ execution."""



"""Includes DictDataFrame class."""
from typing import List, Any, Union, Dict

import numpy as np
import pandas as pd

from organon.fl.core.businessobjects.idataframe import IDataFrame
from organon.fl.core.helpers import string_helper, list_helper


class Np2dDataFrame(IDataFrame):
    """Dictionary based custom DataFrame implementing IDataFrame abstract class."""

    def __init__(self, row_count: int, max_column_count: int, dtype: type):
        super().__init__(row_count=row_count)
        self.max_column_count = max_column_count
        self.dtype = dtype
        self.data_frame: np.ndarray = np.full((row_count, max_column_count), None, dtype=dtype)
        self.column_name_map: Dict[str, int] = {}
        self.empty_column_indices: List[int] = list(range(max_column_count))
        self.column_name_reverse_map: Dict[int, str] = {}
        self.__columns_marked_for_gc = []

    def set_value(self, value, column_name: str, row_index: int):
        index = self.column_name_map[column_name]
        self.data_frame[row_index][index] = value

    def contains_column(self, column_name: str) -> bool:
        return column_name in self.column_name_map

    def try_add(self, column_name: str, data: List[Any] = None, mark_for_garbage_collection: bool = False) -> bool:
        cond1 = string_helper.is_null_or_empty(column_name) or column_name in self.column_name_map
        if cond1:
            raise ValueError("Column name is inadmissible")
        cond2 = not list_helper.is_null_or_empty(data) and len(data) != self.row_count
        if cond2:
            raise ValueError("Data length should be equal to frame-row-count")
        if len(self.empty_column_indices) == 0:
            raise ValueError("Np2dDataFrame has all columns full.")
        new_index = self.empty_column_indices[0]
        self.column_name_map[column_name] = new_index
        self.column_name_reverse_map[new_index] = column_name
        self.empty_column_indices.remove(new_index)
        if data is not None:
            self.data_frame[:, new_index] = np.array(data)
        if mark_for_garbage_collection:
            self.__columns_marked_for_gc.append(column_name)
        return True

    def remove(self, column_names: Union[List[str], str]):
        columns_to_remove = []
        if isinstance(column_names, list):
            columns_to_remove.extend(column_names)
        else:
            columns_to_remove.append(column_names)
        for col in columns_to_remove:
            val = self.column_name_map.pop(col, None)
            if val is not None:
                self.data_frame[:, val] = None
                self.empty_column_indices.append(val)
                self.column_name_reverse_map.pop(val)

    def get_value(self, column_name: str, row_index: int = None):
        index = self.column_name_map[column_name]
        if row_index is not None:
            return self.data_frame[row_index, index]
        return self.data_frame[:, index]

    def collect_garbage_columns(self):
        self.remove(self.__columns_marked_for_gc)
        self.__columns_marked_for_gc.clear()

    def get_column_names(self) -> List[str]:
        return list(self.column_name_map.keys())

    def get_subset_as_pandas_df(self, indices: List[Any] = None, columns: List[str] = None):

        if indices is None and columns is None:
            raise ValueError("Both indices and columns are None")

        if indices is None:
            subset = self.data_frame[:, [self.column_name_map[col] for col in columns]]
        elif columns is None:
            columns = list(self.column_name_map.keys())
            subset = self.data_frame[indices, [self.column_name_map[col] for col in columns]]
        else:
            subset = self.data_frame[indices, [self.column_name_map[col] for col in columns]]
        return pd.DataFrame(subset, columns=columns)

    def copy_as_pandas_df(self):
        column_indices_sorted = sorted(self.column_name_reverse_map.keys())
        columns = []
        for i in column_indices_sorted:
            columns.append(self.column_name_reverse_map[i])
        new_arr = self.data_frame[:, column_indices_sorted]
        return pd.DataFrame(new_arr, columns=columns)

    def rename_columns(self, new_names_dict: Dict[str, str]):
        """
        Renames columns in dataframe
        :param new_names_dict: dictionary mapping old column names to their new names
        """
        new_column_name_map = {}
        new_column_name_reverse_map = {}
        for old_name, new_name in new_names_dict.items():
            col_index = self.column_name_map[old_name]
            new_column_name_map[new_name] = col_index
            new_column_name_reverse_map[col_index] = new_name

        self.column_name_map = new_column_name_map
        self.column_name_reverse_map = new_column_name_reverse_map

    def get_column_name_by_index(self, col_index: int):
        """Returns column name corresponding to given index"""
        return self.column_name_reverse_map[col_index]



"""This module includes enum class 'DqRecordSourceType'"""
from enum import Enum


class DqRecordSourceType(Enum):
    """
    Types of dq record sources.
    """
    DATABASE = 1
    TEXT = 2
    DATA_FRAME = 3
